import boto3
import json
import logging
import random
import threading
import urllib3
import uuid


# AWS Ken Walsh 8-2020
cfn_client = boto3.client('cloudformation')
s3client = boto3.client('s3')
sc_client = boto3.client('servicecatalog')

http = urllib3.PoolManager()
logger = logging.getLogger()
_uu = str(uuid.uuid4())


def gen_surl(bucketname, keyname):
    url = s3client.generate_presigned_url(ClientMethod='get_object', Params={
                                          'Bucket': bucketname, 'Key': keyname})
    return url
    
def sc_deactivat_ver(_newversion, _prodinfo):
    for _product in _prodinfo:
        if _newversion != _product['Name']:
            response = sc_client.update_provisioning_artifact(AcceptLanguage='en', ProductId=_product[
                                                              'ProductId'],    ProvisioningArtifactId=_product['ProvisioningArtifactId'], Active=False, Guidance='DEPRECATED')
    return 0


def sc_create_pa(ProductId, version, description, template):
    IdempotencyToken = "tok%s" % str(int(random.random() * 10000000))
    rep = sc_client.create_provisioning_artifact(AcceptLanguage='en', ProductId=ProductId, Parameters={
        'Name': version,
        'Description': description,
        'Info': {
            "LoadTemplateFromURL": template
        },
        'Type': 'CLOUD_FORMATION_TEMPLATE',
        'DisableTemplateValidation': False
    },
        IdempotencyToken=IdempotencyToken
    )

    return rep


def sc_describe_product(name):
    try:
        rep = sc_client.describe_product_as_admin(
            AcceptLanguage='en', Name=name)
    except Exception as e:
        e = str(e)
        if e.find('ResourceNotFoundException') > -1:
            print(name + ' not found')
            return 'not found'
    ProductId = rep['ProductViewDetail']['ProductViewSummary']['ProductId']
    # print(rep)
    # print(rep['ProvisioningArtifactSummaries'])
    _ret = []
    for _prodversion in rep['ProvisioningArtifactSummaries']:
        _version = {'ProductId': ProductId, 'ProvisioningArtifactId': _prodversion[
            'Id'], 'Name': _prodversion['Name'], 'Description': _prodversion['Description']}
        _ret.append(_version)
    return _ret


def b_put(bucket, key, body):
    srep = s3client.put_object(
        Body=body, Bucket=bucket, Key=key, ContentType='text/json',)
    return srep


def b_get_obj(bucket, bkey):
    object = s3client.get_object(Bucket=bucket, Key=bkey)
    return(object['Body'].read().decode('utf-8'))


def gpar(fjsonpar, key):
    _ret = 'notfound'
    for _p in fjsonpar:
        if key == _p['ParameterKey']:
            _ret = _p['ParameterValue']
    return _ret


def sap_hana_inf(stackName, _newpara, _metadata):
    _template_json = cfn_client.get_template(
        StackName=stackName)['TemplateBody']
    _template_json['Description'] = '%s (fdp-1r56d8at2) v' % _template_json['Description']
    fjsonpara = _template_json['Parameters']
    fjsonpar = cfn_client.describe_stacks(StackName=stackName)[
        'Stacks'][0]['Parameters']
    #return fjsonpara
    for _pp in fjsonpara:
        _val = gpar(fjsonpar, _pp)
        fjsonpara[_pp]['Default'] = _val
        # print(_buf)
    _template_json['Parameters'] = fjsonpara
    #return _template_json['Parameters']
    fin_json = _template_json
    fin_para = fin_json['Parameters']

    for _rp in fin_json['Resources']:
        #print('Checking ' +_rp)
        if 'Parameters' in fin_json['Resources'][_rp]['Properties']:
            for _newp in fin_json['Resources'][_rp]['Properties']['Parameters']:
                if _newp in fin_para:
                    # print('found '+ _newp)
                    fin_json['Resources'][_rp]['Properties'][
                        'Parameters'][_newp] = fin_para[_newp]['Default']
    # update fnsubs
    fin_json['Metadata'] = {}
    fin_json['Conditions'] = {}
    fin_json['Parameters'] = _newpara
    fin_json['Metadata'] = _metadata
    fin_json['Resources']['HANAStack']['Properties']['Parameters']['DomainName'] = {"Ref": "AWS::NoValue"}
    
    
    _hanastack_hold = fin_json['Resources']['HANAStack']
    # check for security group run before
    if _hanastack_hold['Properties']['Parameters']['CreateNewSecurityGroup'] == 'Yes' :
        HANADBSecurityGroupID = cfn_get_output(stackName,'HANAMasterSecurityGroup')
        _hanastack_hold['Properties']['Parameters']['HANADBSecurityGroupID'] = HANADBSecurityGroupID
        _hanastack_hold['Properties']['Parameters']['CreateNewSecurityGroup'] = 'No'
    
    if 'DependsOn'in _hanastack_hold:
        dependson = _hanastack_hold['DependsOn']
        _hanastack_hold_dependson = fin_json['Resources'][dependson]
    
    #check for ExecutePostConfigurationScripts
    execupost = 0
    if 'ExecutePostConfigurationScripts' in fin_json['Resources']:
        _holdExecutePostConfigurationScripts =  fin_json['Resources']['ExecutePostConfigurationScripts']
        _holdExecutePostConfigurationScripts['DependsOn'] = 'HANAStack'
        _holdExecutePostConfigurationScripts['Metadata'] = {}
        _holdExecutePostConfigurationScripts['Properties']['Parameters']['AssetsS3BucketRegion'] = fin_para['AssetsS3RegionCode']['Default']
        #_holdExecutePostConfigurationScripts['Properties']['Parameters']['InstanceIDs'] = {"Fn::GetAtt": ["HANAStack","Outputs.HANAInstanceIdList"]}
        execupost = 1
        
    fin_json['Resources'] = {}
    
    #check for dynamoWorkFlowDB

    if 'DependsOn'in _hanastack_hold:
        fin_json['Resources'][dependson] = _hanastack_hold_dependson 
    if "WorkflowDB" in fin_json['Resources']:
        #table name must start with 'LaunchWizard' for built in role to work
        _randtbuuid = _uu.split('-')[4]
        dydb_tablename = {"Fn::Sub": "LaunchWizard-SC-lz-sapeasy-${AWS::StackName}"}
        
        fin_json['Resources']['WorkflowDB']['Properties']['TableName'] = dydb_tablename
        
    if execupost == 1:
        _holdExecutePostConfigurationScripts['Properties']['Parameters']['DynamoDBTableName'] = dydb_tablename
        fin_json['Resources']['ExecutePostConfigurationScripts'] = _holdExecutePostConfigurationScripts
        
        
   
    
    fin_json['Resources']['HANAStack'] = _hanastack_hold
    fin_json['Resources']['HANAStack']['Properties']['Tags'] =  [{"Key":'LaunchWizardApplicationType',"Value":"SAP"},{"Key": "LaunchWizardResourceGroupID","Value": _uu}]
    #update sc quick parameters
    _sc =  random.randint( 10, 50 )
    fin_json['Parameters']['HANASID']['Default']        = fin_json['Resources']['HANAStack']['Properties']['Parameters']['HANASID'] 
    fin_json['Parameters']['HANAHostname']['Default']   = fin_json['Resources']['HANAStack']['Properties']['Parameters']['HANAHostname'] + str(_sc)
    fin_json['Parameters']['SAPInstanceNum']['Default'] = fin_json['Resources']['HANAStack']['Properties']['Parameters']['SAPInstanceNum'] 
    ##
    fin_json['Resources']['HANAStack']['Properties']['Parameters']['HANASID'] = {"Ref":"HANASID"}
    fin_json['Resources']['HANAStack']['Properties']['Parameters']['HANAHostname']   = {"Ref":"HANAHostname"}
    fin_json['Resources']['HANAStack']['Properties']['Parameters']['SAPInstanceNum'] = {"Ref":"SAPInstanceNum"}
    fin_json['Resources']['HANAStack']['Properties']['Parameters']['UniqueID'] = _uu
    
    
    
    _tmpout = json.dumps(fin_json, indent=3, sort_keys=True)

    for _psrc in fin_para:
        _vvalue = fin_para[_psrc]['Default']
        _sval = '${%s}' % _psrc
        _tmpout = _tmpout.replace(_sval, _vvalue)

    return json.loads(_tmpout)

def cfn_get_output(stackName,outputkey):
    repstacks =response = cfn_client.list_stacks( StackStatusFilter=['CREATE_COMPLETE'])
    _search ='%s-SecurityGroups' % stackName
    ret={}
    for fstacks in repstacks['StackSummaries']:
        if fstacks['StackName'].find(_search) > -1:
            outkeys =  cfn_client.describe_stacks(StackName=fstacks['StackName'])['Stacks'][0]['Outputs']
            for outkey in outkeys:
                if outkey['OutputKey'] == outputkey:
                    ret = outkey['OutputValue']
    return ret   
def cr_sc_product(prodProperties):
    IdempotencyToken = "tok%s" % str(int(random.random() * 10000000))
    
    rep = sc_client.create_product(
        AcceptLanguage='en',
        Name=prodProperties['Name'],
        Owner='MP SC Team',
        Description=prodProperties['Description'],
        Distributor='AWS',
        SupportDescription='AWS',
        SupportEmail='supportaws@awsdemo.com',
        SupportUrl='https://demompscaws.net',
        ProductType='CLOUD_FORMATION_TEMPLATE',
        Tags=[
        {
            'Key': 'LaunchWizardApplicationType',
            'Value': 'SAP'
        },
        {
            'Key': 'LaunchWizardResourceGroupID',
            'Value': _uu
        }
    ],
        ProvisioningArtifactParameters={
            'Name': prodProperties['version'],
            'Description': 'Description ' + prodProperties['version'],
            'Info': {
                "LoadTemplateFromURL": prodProperties['LoadTemplateFromURL']
            },
            'Type': 'CLOUD_FORMATION_TEMPLATE',
            'DisableTemplateValidation': True
        },
        IdempotencyToken=IdempotencyToken
    )
    return rep


def cfnresponse(event, context, responseStatus, responseData, physicalResourceId=None, noEcho=False):
    responseBody = {}
    responseBody['Status'] = responseStatus
    responseBody['Reason'] = 'See the details in CloudWatch Log Stream: ' + \
        context.log_stream_name
    responseBody['PhysicalResourceId'] = physicalResourceId or context.log_stream_name
    responseBody['StackId'] = event['StackId']
    responseBody['RequestId'] = event['RequestId']
    responseBody['LogicalResourceId'] = event['LogicalResourceId']
    responseBody['NoEcho'] = noEcho
    responseBody['Data'] = responseData
    json_responseBody = json.dumps(responseBody)
    headers = {'content-type': '',
               'content-length': str(len(json_responseBody))}
    print("Response body:\n" + json_responseBody)
    try:
        response = http.request(
            'PUT', event['ResponseURL'], body=json_responseBody.encode('utf-8'), headers=headers)
        logger.debug('Status code: ' + response.reason)
    except Exception as e:
        logger.error(
            'cfnresponse(..) failed executing requests.put(..): ' + str(e))


def timeout(event, context):
    logging.error(
        'Execution is about to time out, sending failure response to CloudFormation')
    cfnresponse(event, context, 'FAILED')


def lambda_handler(event, context):
    print(json.dumps(event))
    _ret = {}
    timer = threading.Timer((context.get_remaining_time_in_millis()
                             / 1000.00) - 0.5, timeout, args=[event, context])
    timer.start()
    status = 'SUCCESS'

    try:
        if event['RequestType'] == 'Create':
            _newpara = {"ServerSize": {"Description": "Size", "Default": "r5.8xlarge", "AllowedValues": [
                "r5.8xlarge"], "Type": "String"}, \
                "OwnerEmail": {"Description": "Email", "Default": "sapadmin@mysap.com", "Type": "String"},\
                "HANASID": {"Description": "HANASID",  "Type": "String"},
                "HANAHostname":  {"Description": "HANAHostname",   "Type": "String"},
                "SAPInstanceNum": {"Description": "SAPInstanceNumor HANA System No",  "Type": "String"}
            }
            _metadata = {"AWS::CloudFormation::Interface": {
                    "ParameterGroups": [
                        {
                            "Label": {
                            "default": "Instance Configuration"
                            },
                        "Parameters": [
                            "HANASID",
                            "HANAHostname",
                            "SAPInstanceNum",
                            "ServerSize",
                            "OwnerEmail" 
                        ]}]}}
            _sc_productname = 'SAP-Hana Infrastructure SC Easy'
            _workingbucket = event['ResourceProperties']['HoldingBucket']
            _workingkey = '%ssclwizsap.json' % event['ResourceProperties']['Prefix']
            _dbuf = 'bucket %s key %s' % (_workingbucket,_workingkey)
            print(_dbuf)
            _body = sap_hana_inf(
                event['ResourceProperties']['TemplateName'], _newpara, _metadata)
            #return _body 
            # Check if the product exist
            _prodinfo = sc_describe_product(_sc_productname)
            b_put(_workingbucket, _workingkey, json.dumps(_body, indent=4))
            _returl = gen_surl(_workingbucket,_workingkey)
            _ret['uu'] =  _returl
            #return _ret
            
            
            if _prodinfo == 'not found':
                b_put(_workingbucket, _workingkey, json.dumps(_body, indent=4))
                _s3cft = 'https://%s.s3.amazonaws.com/%s' % (
                    _workingbucket, _workingkey)
                version = 'Latest 1.0'
                description = 'Description %s' % version
                template = _s3cft
                prodProperties = {'version': version, 'Name': _sc_productname,
                                  'Description': 'SC SAP HANA Easy Deploy', 'LoadTemplateFromURL': _s3cft}

                prodinfo = cr_sc_product(prodProperties)
                prodProperties['ProductId'] = prodinfo['ProductViewDetail']['ProductViewSummary']['ProductId']
                prodProperties['ProvisioningArtifactId'] = prodinfo['ProvisioningArtifactDetail']['Id']
                _ret = prodProperties

            else:
                b_put(_workingbucket, _workingkey, json.dumps(_body, indent=4))
                _s3cft = 'https://%s.s3.amazonaws.com/%s' % (
                    _workingbucket, _workingkey)
                _prodinfo = sc_describe_product(_sc_productname)
                _versions = len(_prodinfo) + 1
                ProductId = _prodinfo[0]['ProductId']
                version = 'Latest %s.0' % _versions
                description = 'Decription %s' % version
                template = _s3cft
                pa_rep = sc_create_pa(ProductId, version,
                                      description, template)
                _rr = {'ProductId': ProductId, 'version': version,
                       'ProvisioningArtifactId': pa_rep['ProvisioningArtifactDetail']['Id']}
                _ret = _rr
                # Deactivate old version
                _prodinfo = sc_describe_product(_sc_productname)
                sc_deactivat_ver(version, _prodinfo)
        else:
            timer.cancel()
            cfnresponse(event, context, status, _ret)
    except Exception as e:
        print(e)
        logging.error('Exception: %s' % e, exc_info=True)
        status = 'FAILED'
    finally:
        print(json.dumps(_ret))
        timer.cancel()
        cfnresponse(event, context, status, _ret)
